#!/bin/sh  
#\
exec tclsh "$0" "$@"

# dirtree.tcl
#	demonstrate using a tree package
#
# Copyright (c) 1998 Clif Flynt 
#
# This program is free for use in non-commercial applications, provided 
# that the copyright information is retained.  It may be licensed from
# Flynt Consulting Services for use in commercial applications.    
#
#  This code is described in "Tcl/Tk for Real Programmers", 
#  published by Academic Press Professional
#  ISBN: 0122612051 
#



source tree.tcl

########################################################################
# proc addDirToTree
# 	Add a directory to a node.  
#	Create children for each directory entry, and set the name and
#	  type of each child.
#
# Arguments:
#	parent		The parent node for this directory.
#	directory	The directory to add to the tree.
#
# Results:
#	The Tree is made longer by the number of entries in this directory.

proc addDirToTree {parent directory} {

    #
    # If this isn't a directory, it has no subordinates.
    #

    if {[file type $directory] != "directory"} {
        error "Target for adding a directory to tree must *be* a directory"
    }

    #
    # If the parent directory hasn't been updated with name and type,
    #   do so now.
    #
    if {[catch {treeGetNodeValue $parent name}]} {
        treeSetNodeValue $parent name $directory
        treeSetNodeValue $parent type directory
	}
    
    #
    # An empty or unreadable directory will return a fail for the glob command
    # If the directory can be read, the list of names will be in fileList.
    #

    set fail [catch {glob $directory/*} fileList]

    if {$fail} {
        return;
    }
    
    #
    # Loop through the names in fileList
    #
    foreach name $fileList {
        set node [treeCreateChild $parent]
        treeSetNodeValue $node name $name
  	treeSetNodeValue $node type [file type $name]
    }
}

#
# Create a tree structure
#
set base [treeCreate ]

#
# Add the initial directory and it's directory entries to the tree structure
#
addDirToTree $base [file nativename /]

#
# If any children are directories, add their entries to the tree under as
#  sub-children.
#
foreach entry [treeGetChildList $base] {
    if {[treeGetNodeValue $entry type] == "directory"} {
        addDirToTree $entry [treeGetNodeValue $entry name]
    }
}

#
# Print out the tree in the form:
# Dir1
#    Dir1/Entry1
#    Dir1/Entry2
# Dir2
#    Dir2/Entry3
#    Dir2/Entry4
#

foreach entry [treeGetChildList $base] {
    puts "[treeGetNodeValue $entry name]"
    if {[treeGetNodeValue $entry type] == "directory"} {
        foreach sub [treeGetChildList $entry] {
	    puts "    [treeGetNodeValue $sub name]"
	}
    }
}
