# showTree.tcl
#	Displays a tree in a canvas.  Will load moen or nestedbox layout
#   techniques as required.  Uses the Tree as a data container for lots of
#   stuff.
#
# Copyright (c) 1998 Clif Flynt 
#
# This program is free for use in non-commercial applications, provided 
# that the copyright information is retained.  It may be licensed from
# Flynt Consulting Services for use in commercial applications.    
#
#  This code is described in "Tcl/Tk for Real Programmers", 
#  published by Academic Press Professional
#  ISBN: 0122612051 
#

# Todo: nodes carry width/height/spacing info, instead of xoff/yoff
# 
set CanvasTree_version [lindex {$RCSfile: showTree.tcl,v $ $Revision: 1.25 $} 4]


package provide canvastree 1.0

################################################################
# proc newCanvasTree {} --
#     Creates a new CanvasTree
#
# Arguments
#     NONE
# Results
#     Adds a new tree to the Tree table
#     Sets defaults
#     Creates a first node and returns it.

proc newCanvasTree {} {
    set newNode [canvastree::newTreeObject]
    proc $newNode {args} "
	return \[canvastree::CanvasTreeObjProc {$newNode} \$args\]
    "
    canvastree::CanvasTreeDefaults $newNode

    $newNode set -tree layout horizontal
    $newNode set -tree renderProc defaultRenderNode
    $newNode set -tree parent_distance 5

    return $newNode
}

namespace eval canvastree {

package require treeObject
package require moenlayout
package require nestedbox

if {[info exists auto_index(newTreeObject)]} {
    eval $auto_index(newTreeObject)
}


if {[info exists auto_index(::moenlayout::moenLayout)]} {
    eval $auto_index(::moenlayout::moenLayout)
    namespace import moenlayout::*
}

if {[info exists auto_index(::nestedbox::BoxLayout)]} {
    eval $auto_index(::nestedbox::BoxLayout)
    namespace import nestedbox::*
}

#
# Defined Node fields:
#
# displayText :	A list of strings to display
# height	   : set by rendering Proc
# width		   : set by rendering Proc

# Defined Node/Tree fields
#
# renderProc  : A user defined rendering proc
# layout      : vertical/horizontal - default horizontal
# yDesiredPosition : The position this node *should* occupy
# xDesiredPosition : This value is used to calculate children positions, etc.
# lineWidth	   : If != 0, draw lines between nodes in renderProc

# xPad		   : Horizontal padding between Nodes
# xIpad		   : Horizontal padding between text bounding box & rectangle
# yPad, yIpad	   : Vertical

# Font		   : A font descriptor

#
# Node Fields used for placement algorithm - 
#   Sven Moen, Brown U, IEEE Software July, 1990, P21
#
# contour.lower.head  : Poly lines that define the bounding polygons.
# contour.lower.tail  : Saved as Xoffset1, Yoffset1, Xoffset2, Yoffset2, ...
# contour.upper.head
# contour.upper.tail
#
# Xoffset, Yoffset	      : offset of this node from parent in pixels.
#
# Tree Fields used for placement algorithm
#  parent_distance : Distance between nodes


########################################################################
# proc CanvasTreeObjProc {node methodArgs} --
#    Primary method for tree objects.  Parses out the subcommand and
#    calls the appropriate tree.tcl proc.
#
# Arguments:
#    node:	  The internal name for this node.
#    methodArgs:  A list, the first member is the subcommand name,
#		  subsequent (optional) members are arguments to be
#		  passed to the tree.tcl command.
#
# Results:
#    Calls other procedures which may produce output, create new objects,
#	delete objects, etc.

proc CanvasTreeObjProc {node methodArgs} {

    set subCommand [lindex $methodArgs 0]
    set cmdArgs [lrange $methodArgs 1 end]

    set methods {display}

    switch -- $subCommand {
    {display} {
	    # Can be called with -r for recursive, else displays
	    #  A single node at declared location
	    #  args : canvasName ?-r?
	    eval displayNode $node $cmdArgs
	}
    {add} {
	    # Add a child node to this node.
	    set newNode [treeObject::treeCreateChild $node]
	    namespace eval :: "proc $newNode {args} \" \
		return \\\[canvastree::CanvasTreeObjProc {$newNode} \\\$args]\"	    "
	    CanvasTreeDefaults $newNode
	    return $newNode
	}
    {default} {
	    return [treeObject::TreeObjProc $node $methodArgs]
	}
    }
}


################################################################
# proc CanvasTreeDefaults {node} --
#   Sets the defaults for a CanvasTree
# Arguments
#   node:	The node to set defaults for
# Results
#   No valid return.
# 
proc CanvasTreeDefaults {node} {
    $node set -node height 0
    $node set -node width 0
    $node set -node contour.upper.tail ""
    $node set -node contour.upper.head ""
    $node set -node contour.lower.tail ""
    $node set -node contour.lower.head ""
    if {[catch {$node get xpad}]} {
        $node set -node xpad 10
    }
    if {[catch {$node get ypad}]} {
        $node set -node ypad 10
    }
    $node set -node xIpad 10
    $node set -node yIpad 10
    if {[catch {$node get lineWidth}]} {
        $node set -node lineWidth	1
    }

}

################################################################
# proc CanvasTreeSetSize {node} --
#     Sets the height & width for a node.
#     Does this by creating an invisible canvas, drawing the node, and
#     destroying the canvase.
# Arguments
#     node:	The node to determine the size of
# Results
#     Sets the height and width fields for the node
# 

proc CanvasTreeSetSize {node} {
    canvas .xxyyzz -height 500 -width 500
    $node set -node xDesiredPosition 250
    $node set -node yDesiredPosition 50
    eval [$node get renderProc] $node .xxyyzz 
    destroy .xxyyzz
}


################################################################
# proc displayNode {node cvs {recurs ""}} --
#   Calls the defined renderProcs and draw lines if req'd
# Arguments
#  node:	The node to draw
#  cvs:		The canvas to draw on
#  ? -r ?	If -r present, draw recursively all subordinate nodes.
#
# Results
#   Fills a canvas with junk
# 
proc displayNode {node cvs {recurs ""}} {

    eval [$node get renderProc] $node $cvs 
    
    set pbb [$cvs bbox $node.box]
  if {[string match [string tolower [$node get layout]] "horizontal"]} {
      set yp [expr ([lindex $pbb 1] +[lindex $pbb 3])/2]
      set xp [lindex $pbb 2]
  } else {
      set xp [expr ([lindex $pbb 0] +[lindex $pbb 2])/2]
      set yp [lindex $pbb 3]
  }

    if {[string match $recurs "-r"]} {
	foreach c [$node childList] {
	    displayNode $c $cvs -r
    if {[set lineWidth [$c get lineWidth]]} {	    
    set cbb [$cvs bbox $c.box]
  if {[string match [string tolower [$node get layout]] "horizontal"]} {
      set yc [expr ([lindex $cbb 1] +[lindex $cbb 3])/2]
      set xc [lindex $cbb 0]
  } else {
      set xc [expr ([lindex $cbb 0] +[lindex $cbb 2])/2]
      set yc [lindex $cbb 1]
  }
  set t [$cvs create line $xp $yp $xc $yc -tag $c -width $lineWidth]
  $cvs addtag $c.line withtag $c
}
	}
    }

}

################################################################
# proc defaultRenderNode {node cvs args} -
#   The default node drawer - 
#	Places text with a simple rectangle around it.
# Arguments
#   node:	The node to render
#   cvs	:	The canvas to render on
#   args:	optional arguments
# Results
#   Lots of lines and such.
# 
proc defaultRenderNode {node cvs args} {
    eval drawText $node $cvs $args

    set bb [textBB $node $cvs ]

    set t [eval $cvs create rectangle $bb $args]
    $cvs addtag $node withtag $t 
    $cvs addtag $node.box withtag $t 

    $node set -node height [expr [lindex $bb 3] - [lindex $bb 1]]
    $node set -node width [expr [lindex $bb 2] - [lindex $bb 0]]
}

################################################################
# proc drawText {node cvs args} --
#   Draws the text on the canvas
# Arguments
#   node:	The node to get the text from
#   cvs:	The canvas to draw to
#   args:	Optional args.
# Results
#   Draws lines of text (hardcoded font)
#   Sets tags for $node and $node.txt

proc drawText {node cvs args} {
    set txt [$node get displayText]
    set xDesiredPosition [$node get xDesiredPosition]
    set yDesiredPosition [$node get yDesiredPosition]
    
    if {[catch {$node get Font} font]} {
        set font [list helvetica 12 bold]
    }

    set fontHeight [lindex [font metrics $font ] 5]

    foreach tt $txt {
	set t [eval $cvs create text $xDesiredPosition $yDesiredPosition -text \$tt\
	  -font {$font} $args]
	$cvs addtag $node.txt withtag $t 
	$cvs addtag $node withtag $t
	incr yDesiredPosition $fontHeight
    }
}

################################################################
# proc textBB {node cvs } --
#   Returns the bounding box for a nodes text
# Arguments
#  node:	The node identifier
#  cvs:		The canvas with text
# Results
#   Returns the bounding box, with internal padding added.

proc textBB {node cvs } {

    set bb [$cvs bbox $node.txt]

    set xpad [$node get xIpad]
    set ypad [$node get yIpad]

    set xl [lindex $bb 0]
    incr xl -$xpad
    set bb [lreplace $bb 0 0 $xl]

    set yt [lindex $bb 1]
    incr yt -$ypad
    set bb [lreplace $bb 1 1 $yt]

    set xr [lindex $bb 2]
    incr xr $xpad
    set bb [lreplace $bb 2 2 $xr]

    set yb [lindex $bb 3]
    incr yb $ypad
    set bb [lreplace $bb 3 3 $yb]
    return $bb
}

################################################################
# proc raisedTextRenderNode {node cvs args} -
#   Makes a raised rectangle with text
# Arguments
#   node:	The node to render
#   cvs	:	The canvas to render on
#   args:	optional arguments
# Results
#   Lots of lines and such in pretty boxes
# 
# 
proc raisedTextRenderNode {node cvs args} {
    eval drawText $node $cvs $args
    set bb [textBB $node $cvs ]
    
    raisedRect $cvs [lindex $bb 0] [lindex $bb 1] [expr [lindex $bb 2] -\
      [lindex $bb 0]] [expr [lindex $bb 3] - [lindex $bb 1]] -tag $node.box
    $cvs lower $node.box

    $node set -node height [expr [lindex $bb 3] - [lindex $bb 1]]
    $node set -node width [expr [lindex $bb 2] - [lindex $bb 0]]
}

################################################################
# proc raisedRect {c x y w h args} --
#   Makes a raised rectangle around a defined area. 
# Arguments
#   c:	canvas
#   x:  left x position
#   y:	top y position
#   w:  width of center rectangle
#   h:  height of center rectangle
#   args: optional args.
# Results
#   Draws a rectangle around the defined coords.  Args can be used to tag
#     the raised rectangle.

proc raisedRect {c x y w h args} {
set wid 4
set lwid 3

incr x  
incr y

incr x -$wid

set xs $x; incr xs ;
set ys $y; incr ys ;

incr h $wid
incr w $wid


eval "$c create rectangle $xs $ys [expr $x+$w] [expr $y+$h] \
    -width $wid -fill gray80 -outline black $args"
    
    eval "$c create polygon $x [expr $y+$h] $x $y [expr $x+$w] $y \
       [expr $x +$w - $lwid] [expr $y+$lwid] [expr $x+$lwid] [expr $y+$lwid] \
       [expr $x+$lwid] [expr $y+$h-$lwid] \
       -width 0 -fill gray95 -outline gray95  $args"

}

proc testShow {cvs node} {
    set xDesiredPosition [$node get xDesiredPosition]
    set yDesiredPosition [$node get yDesiredPosition]
    $node set -node xActualPosition $xDesiredPosition


    # xDesiredPosition is the "official" position of a node - for other 

    #  nodes to use to figure out where they should be.

    # xActualPosition is the actual position of a node - for nodes 

    #  and lines that need to know exactly where something is.

    set fail [catch {$node parent} parent]
    if {$fail} {
	set parent ""
    }

    # if this is the root node, we should try to center it.
    if {$parent == ""} {
	set xoff [$node get xoff]
	set childCount [llength [$node childList]]
	# If even number of children, move xDesiredPosition left by an xoff/2
	if {[expr $childCount %2] == 0} {
	    set xDesiredPosition [expr $xDesiredPosition - $xoff/2]
	    $node set -node xActualPosition $xDesiredPosition
	}
    }

    array set offs {0 -4 1 -4 2 2 3 1}
    set fontHeight [lindex [font metrics [list arial 12 bold]] 5]

    set txtlst [$node get displayText]
    foreach txt $txtlst {
	set id [$cvs create text $xDesiredPosition $yDesiredPosition\
	  -text $txt -font [list arial 12 bold] -anchor n -tag $node]
	$cvs addtag $node.txt withtag $node
	incr yDesiredPosition $fontHeight
    }
    set bb [$cvs bbox $node.txt]

    for {set i 0} {$i < 4} {incr i} {
	set j [lindex $bb $i]
	incr j $offs($i)
	set bb [lreplace $bb $i $i $j]
    }
    set wd [expr [lindex $bb 2] - [lindex $bb 0]]
    set ht [expr [lindex $bb 3] - [lindex $bb 1]]

    $node set -node height $ht
    $node set -node width $wd
    $node set -node yoff [expr $ht + 25]


    #    raisedRect $cvs [lindex $bb 0] [lindex $bb 1] $wd $ht -tag $node.box

    #    $cvs addtag $node withtag $node.box

    $cvs raise $node.txt

    if {$parent != ""} {

	#        set px [expr [$parent get xDesiredPosition] + ([$parent get\
	  width]/2)]
	set px [$parent get xActualPosition]
	set py [expr [$parent get yDesiredPosition] + [$parent get height]\
	  - $fontHeight]
	incr py 11
	set xp $xDesiredPosition
	set yp [expr $yDesiredPosition - $ht]

	# $cvs create line $xp $yp $px $py -fill black -width 2
    }
}

################################################################
# proc setDesiredPositions {node} --
#   Sets the absolute positions for nodes based on Xoffset, Yoffset
# Arguments
#   node:	The base node to work from.  
#	xDesiredPosition and yDesiredPosition must be set for this node.
# Results
# 	All nodes get absolute coords defined
# 
proc setDesiredPositions {node} {

    set myChildList [$node childList]

    if {[llength $myChildList] == 0} {
	return
    }
    set ch0 [lindex $myChildList 0]
    set myChildList [lreplace $myChildList 0 0]
    set ch0w [$ch0 get width]

    set ox [expr [$node get xDesiredPosition] + [$ch0 get Xoffset] + \
    	   $ch0w/2 - [$node get width]/2]
    set oy [expr [$node get yDesiredPosition] + [$ch0 get Yoffset]]

    $ch0 set -node xDesiredPosition $ox
    $ch0 set -node yDesiredPosition $oy

    foreach c $myChildList {
	set nx [expr $ox + [$c get Xoffset] - $ch0w/2 + [$c get width]/2]
	set ny [expr $oy + [$c get Yoffset] ]
	$c set -node xDesiredPosition $nx
	$c set -node yDesiredPosition $ny
# puts "$c: nx: $nx set ox: $ox set ch0w: $ch0w - cwid: [$c get width] off: [$c get Xoffset]"
	set ox $nx
	set oy $ny
	set ch0w [$c get width]
    }

    set myChildList [$node childList]
    foreach c $myChildList {
	setDesiredPositions $c
    }

}

################################################################
# proc ShowTreeTest {}--
#    More complex test of early system.
# Arguments
#   None.
# 
# Results
#   Generates display
# 
proc ShowTreeTest {} {
    set base [newCanvasTree]
    $base set -node displayText [list "Global" "" "all procedure names"\
      "all command names" "all global variables" "exported namespace procs and\
      variables"]
    $base set -node xoff 264
    $base set -node yoff 40
    $base set -tree renderProc testShow

    set ch1 [$base add]
    $ch1 set -node displayText [list "proc stack1" "" "can access local\
      (stack1) scope" "can access global scope" "can access exported ::stack1\
      scope"]
    set ch1_1 [$ch1 add]
    $ch1_1 set -node displayText [list "proc stack2" "" "can access local\
      (stack2) scope" "can access global scope" "can access proc stack1 scope"\
      "can access exported ::stack1 scope"]
    set ch1_2 [$ch1_1 add]
    $ch1_2 set -node displayText [list "proc stack3" "" "can access local\
      (stack3) scope" "can access global scope" "can access proc stack1 scope"\
      "can access proc stack2 scope" "can access exported ::stack1 scope"]
    $ch1 set -node xoff 264


    #  set ch2 [$base add]

    #  $ch2 set -node displayText "proc b1"

    #  $ch2 set -node displayText [list "proc b1" "can access local scope"\
      "can access global scope"]

    #  set ch2_1 [$ch2 add]

    #  $ch2_1 set -node displayText [list "proc b2" "can access local scope"\
      "can access global scope" "can access proc b1 scope"]

    #  $ch2 set -node xoff 200

    set nam1 [$base add]
    $nam1 set -node displayText [list "namespace  ::space1" "" "creates scope\
      space1" "can access space1 variables" "can access space1 procedures"\
      "can access global scope" "" "Can export variables and procedure names"]
    set nam1_1 [$nam1 add]
    $nam1_1 set -node displayText [list "proc  ::space1::proc1" "" "can access\
      local (proc1) scope" "can access namespace  ::space1 scope" "can access\
      global scope"]
    $nam1 set -node xoff 264
    $nam1_1 set -node layout vertical

    set nam1_2 [$nam1 add]
    $nam1_2 set -node displayText [list "proc  ::space1::proc2" "" "can access\
      local (proc2) scope" "can access namespace  ::space1 scope" "can access\
      global scope"]

    #  $nam1_2 set -node xDesiredPosition 250

    #  $nam1_2 set -node yDesiredPosition 350
    $nam1_2 set -node layout vertical


    #  set box [$nam1 add]

    #  $box set -node renderProc blockShow

    #  $box set -node displayText [list "namespace  ::space1"\
      "namespace  ::space1 variables" "namespace  ::space1 procedures"]



    #  set nam2 [$base add]

    #  $nam2 set -node displayText [list "namespace  ::B1" "can access\
      namespace  ::B1 scope" "can access global scope"]

    #  set nam2_1 [$nam2 add]

    #  $nam2_1 set -node displayText [list "namespace  ::B2" "can access\
      namespace  ::B2 scope" "can access global scope"]

    #  $nam2 set -node xoff 264

    set cht 500
    set cwd 620
    canvas .c -height $cht -width $cwd -bg white
    pack .c
    .c create text 325 10 -text "Tcl Variable Scope" -font [list arial 14 bold]

    showCanvasTree $base .c -x 450 -y 40

    CanvasTreeSetBBOX $base .c


    #  set bb_base [$base get BBOX]

    #  .c create rectangle [lindex $bb_base 0] [lindex $bb_base 1]  ;#     \
      [lindex $bb_base 2] [lindex $bb_base 3] -outline red

    #

    #  set bb_base [$ch1 get BBOX]

    #  .c create rectangle [lindex $bb_base 0] [lindex $bb_base 1]  ;#     \
      [lindex $bb_base 2] [lindex $bb_base 3] -outline green

    #

    #  set bb_base [$nam1 get BBOX]

    #  .c create rectangle [lindex $bb_base 0] [lindex $bb_base 1]  ;#     \
      [lindex $bb_base 2] [lindex $bb_base 3] -outline blue

    #

    #  set bb_base [$nam1_1 get BBOX]

    #  .c create rectangle [lindex $bb_base 0] [lindex $bb_base 1]  ;#     \
      [lindex $bb_base 2] [lindex $bb_base 3] -outline orange

    button .b1 -text "print" -command ".c postscript -file scope.ps\
      -pageheight [expr $cht/65]i -pagewidth [expr $cwd/65]i -x 1 -y 1\
      -height $cht -width $cwd"
    button .b2 -text "quit" -command "exit"
    pack .b1
    pack .b2
}


# ShowTreeTest


################################################################
# proc drawContour {cvs node} --
#   Draw the contours for a node.
# Arguments
#   cvs:	The canvas to draw to
#   node:	node to draw contours for
# Results
#   displays on screen.
# 
proc drawContour {cvs node} {

    set color(contour.lower.head) red
    set color(contour.lower.tail) yellow
    set color(contour.upper.head) green
    set color(contour.upper.tail) blue

    foreach id [list contour.lower.head contour.lower.tail contour.upper.head\
      contour.upper.tail] {
	set offsets [$node get $id]
    if {[string match [string tolower [$node get layout]] "horizontal"]} {
	set ox [expr [$node get xDesiredPosition] - ([$node get width] /2) -\
	  [$node get xpad]]
	if {[string first "upper" $id] > 0} {
	    set oy [expr [$node get yDesiredPosition] - 10]
	} else {
	    set oy [expr [$node get yDesiredPosition] + [$node get height]]
	}
	} else {
	  set oy [expr [$node get yDesiredPosition] - 20]
	if {[string first "upper" $id] > 0} {
	set ox [expr [$node get xDesiredPosition] + ([$node get width] /2) +\
	  [$node get xpad]]
	} else {
	set ox [expr [$node get xDesiredPosition] - ([$node get width] /2) -\
	  [$node get xpad]]
	}
	}

	for {set x 0; set y 1} {$y < [llength $offsets]} {incr x 2; incr y 2} {
	    set nx [expr $ox + [lindex $offsets $x]]

	    set ny [expr $oy + [lindex $offsets $y]]

	    $cvs create line $ox $oy $nx $ny -fill $color($id)
	    set ox $nx
	    set oy $ny
	}
    }
}

################################################################
# proc dumpPoly {node} --
#    Prints out the contents of the polylines for a node
# Arguments
#   node: 	The node to output
# Results
#   Displays to stdout.
# 
proc dumpPoly {node} {
    foreach side [list upper lower] {
	foreach end [list head tail] {
	    set id contour.$side.$end
	    puts "$node: $id [$node get $id]"
	}
    }
}

}
